#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <random>
#include <unordered_set>
#include <ctime>
#include <utility>
#include <cstdlib>
#include <deque>
#include <queue>
#include <bitset>

using namespace std;

const int INF = 1e9;
const int UNVISITED = 0;
const int NMAX = 160;
const int PMAX = 11;

struct State
{
    int r, c;
    int mask;
    State(int _r, int _c, int _m) : r(_r), c(_c), mask(_m) {}
};

int typeCell[NMAX][NMAX], switchIndex[NMAX][NMAX], switchMask[NMAX][NMAX];
bitset<1 << PMAX> visited[NMAX][NMAX];

bool isLit(int r, int c, int mask)
{
    if (typeCell[r][c] == 2)
        return false;
    if (typeCell[r][c] == 3)
        return true;

    int bs = typeCell[r][c];
    int sm = switchMask[r][c];
    int toggles = __builtin_popcount(mask & sm);
    int finalState = bs ^ (toggles & 1);
    return (finalState == 1);
}

int main()
{
    freopen("escapelight.in", "r", stdin);
    freopen("escapelight.out", "w", stdout);

    int N, M;
    cin >> N >> M;

    int switchCount = 0;
    vector<pair<int, int>> switchesCoord;

    for (int r = 0; r < N; r++)
    {
        for (int c = 0; c < M; c++)
        {
            switchIndex[r][c] = -1;
            cin >> typeCell[r][c];
            if (typeCell[r][c] == 3)
            {
                switchIndex[r][c] = switchCount;
                switchCount++;
                switchesCoord.push_back({r, c});
            }
        }
    }

    int x_start, y_start, x_stop, y_stop;
    cin >> x_start >> y_start >> x_stop >> y_stop;
    x_start--;
    y_start--;
    x_stop--;
    y_stop--;

    int K, P;
    cin >> K >> P;

    for (int i = 0; i < K; i++)
    {
        int xs, ys, xl, yl;
        cin >> xs >> ys >> xl >> yl;
        xs--;
        ys--;
        xl--;
        yl--;
        int idx = switchIndex[xs][ys];
        switchMask[xl][yl] |= (1 << idx);
    }

    if (!isLit(x_start, y_start, 0))
    {
        cout << -1 << "\n";
        return 0;
    }

    deque<pair<State, int>> dq;
    visited[x_start][y_start][0] = 1;
    dq.push_back({State(x_start, y_start, 0), 1});

    int dir[4][2] = {{1, 0}, {-1, 0}, {0, 1}, {0, -1}};

    while (!dq.empty())
    {
        int r = dq.front().first.r;
        int c = dq.front().first.c;
        int mask = dq.front().first.mask;
        int currentDist = dq.front().second;
        dq.pop_front();

        if (r == x_stop && c == y_stop)
        {
            cout << currentDist << "\n";
            return 0;
        }

        if (typeCell[r][c] == 3)
        {
            int idx = switchIndex[r][c];
            int newMask = mask ^ (1 << idx);
            if (visited[r][c][newMask] == UNVISITED)
            {
                visited[r][c][newMask] = 1;
                dq.push_front({State(r, c, newMask), currentDist});
            }
        }

        for (int i = 0; i < 4; i++)
        {
            int nr = r + dir[i][0];
            int nc = c + dir[i][1];
            if (nr < 0 || nr >= N || nc < 0 || nc >= M)
                continue;

            if (isLit(nr, nc, mask))
            {
                if (visited[nr][nc][mask] == UNVISITED)
                {
                    visited[nr][nc][mask] = 1;
                    dq.push_back({State(nr, nc, mask), currentDist + 1});
                }
            }
        }
    }

    cout << -1 << "\n";
    return 0;
}